;;; custom-load.el --- automatically extracted custom dependencies

;;; Code:

(autoload 'custom-add-loads "cus-load")

(custom-add-loads 'scheme '("semantic-scm"))
(custom-add-loads 'texinfo '("document"))
(custom-add-loads 'speedbar '("semantic-sb"))
(custom-add-loads 'document '("document-vars" "document"))
(custom-add-loads 'semantic-imenu '("semantic-imenu"))
(custom-add-loads 'semantic-faces '("senator"))
(custom-add-loads 'senator '("senator"))
(custom-add-loads 'lisp '("working"))
(custom-add-loads 'semanticdb '("semanticdb"))
(custom-add-loads 'semantic '("semantic-util-modes" "semanticdb" "document" "semantic-imenu" "semantic-util" "semantic" "senator"))
(custom-add-loads 'working '("working"))
(custom-add-loads 'imenu '("semantic-imenu"))
(custom-add-loads 'c '("semantic-c"))

;;; custom-load.el ends here
