;;; DO NOT MODIFY THIS FILE
(if (featurep 'semantic-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "semantic/_pkg.el")

(package-provide 'semantic :version 1.18 :author-version "1.4.2" :type 'regular)

;;;***

;;;### (autoloads (semantic-analyze-current-context) "semantic-analyze" "semantic/semantic-analyze.el")

(autoload 'semantic-analyze-current-context "semantic-analyze" "\
Analyze the current context at POSITION.
If called interactively, display interesting information about POSITION
in a separate buffer.
Returns an object based on symbol `semantic-analyze-context'." t nil)

;;;***

;;;### (autoloads (semantic-bnf-mode) "semantic-bnf" "semantic/semantic-bnf.el")

(defalias 'bnf-mode 'semantic-bnf-mode)

(autoload 'semantic-bnf-mode "semantic-bnf" "\
Initialize a buffer for editing BNF code." t nil)

;;;***

;;;### (autoloads (semantic-default-c-setup) "semantic-c" "semantic/semantic-c.el")

(autoload 'semantic-default-c-setup "semantic-c" "\
Set up a buffer for semantic parsing of the C language." nil nil)

;;;***

;;;### (autoloads (semantic-cb-speedbar-mode) "semantic-cb" "semantic/semantic-cb.el")

(autoload 'semantic-cb-speedbar-mode "semantic-cb" "\
Bring speedbar up, and put it into Class Browser mode.
This will use the Class Browser logic applied to the current Semantic
project database to build the available relations.  The structure of
the class hierarchy can then be navigated using traditional speedbar
interactions." t nil)

;;;***

;;;### (autoloads (semantic-chart-nonterminal-complexity-token semantic-chart-database-size semantic-chart-nonterminals-by-token) "semantic-chart" "semantic/semantic-chart.el")

(autoload 'semantic-chart-nonterminals-by-token "semantic-chart" "\
Create a bar chart representing the number of nonterminals for a token.
Each bar represents how many toplevel nonterminal in BUFFER-OR-STREAM
exist with a given token type.  See `semantic-symbol->name-assoc-list'
for tokens which will be charted." t nil)

(autoload 'semantic-chart-database-size "semantic-chart" "\
Create a bar chart representing the size of each file in semanticdb.
Each bar represents how many toplevel nonterminals in BUFFER-OR-STREAM
exist in each database entry." t nil)

(autoload 'semantic-chart-nonterminal-complexity-token "semantic-chart" "\
Create a bar chart representing the complexity of some tokens.
Complexity is calculated for tokens with a token of SYMBOL.  Each bar
represents the complexity of some nonterminal in BUFFER-OR-STREAM.
Only the most complex items are charted." t nil)

;;;***

;;;### (autoloads (semantic-default-elisp-setup) "semantic-el" "semantic/semantic-el.el")

(autoload 'semantic-default-elisp-setup "semantic-el" "\
Setup hook function for Emacs Lisp files and Semantic." nil nil)

;;;***

;;;### (autoloads (semantic-create-imenu-index) "semantic-imenu" "semantic/semantic-imenu.el")

(autoload 'semantic-create-imenu-index "semantic-imenu" "\
Create an imenu index for any buffer which supports Semantic.
Uses the output of the Semantic Bovinator to create the index.
Optional argument STREAM is an optional stream of tokens used to create menus." nil nil)

;;;***

;;;### (autoloads (semantic-default-java-setup) "semantic-java" "semantic/semantic-java.el")

(autoload 'semantic-default-java-setup "semantic-java" "\
Set up a buffer for semantic parsing of the Java language." nil nil)

;;;***

;;;### (autoloads (semantic-default-make-setup) "semantic-make" "semantic/semantic-make.el")

(autoload 'semantic-default-make-setup "semantic-make" "\
Set up a Makefile buffer for parsing with semantic." nil nil)

;;;***

;;;### (autoloads (semantic-default-scheme-setup) "semantic-scm" "semantic/semantic-scm.el")

(autoload 'semantic-default-scheme-setup "semantic-scm" "\
Setup hook function for Emacs Lisp files and Semantic." nil nil)

;;;***

;;;### (autoloads (semantic-default-skel-setup) "semantic-skel" "semantic/semantic-skel.el")

(autoload 'semantic-default-skel-setup "semantic-skel" "\
Set up a buffer for semantic parsing of the skeleton language." nil nil)

;;;***

;;;### (autoloads (semantic-default-texi-setup) "semantic-texi" "semantic/semantic-texi.el")

(autoload 'semantic-default-texi-setup "semantic-texi" "\
Set up a buffer for parsing of Texinfo files." nil nil)

;;;***

;;;### (autoloads (semantic-summary-mode global-semantic-summary-mode semantic-auto-parse-mode global-semantic-auto-parse-mode semantic-show-unmatched-syntax-mode global-semantic-show-unmatched-syntax-mode semantic-show-dirty-mode global-semantic-show-dirty-mode) "semantic-util-modes" "semantic/semantic-util-modes.el")

(autoload 'global-semantic-show-dirty-mode "semantic-util-modes" "\
Toggle global use of option `semantic-show-dirty-mode'.
If ARG is positive, enable, if it is negative, disable.
If ARG is nil, then toggle." t nil)

(defcustom global-semantic-show-dirty-mode nil "*If non-nil, enable global use of show-dirty mode." :group 'semantic :type 'boolean :require 'semantic-util-modes :initialize 'custom-initialize-default :set (lambda (sym val) (global-semantic-show-dirty-mode (if val 1 -1))))

(autoload 'semantic-show-dirty-mode "semantic-util-modes" "\
Minor mode for highlighting dirty tokens.
With prefix argument ARG, turn on if positive, otherwise off.  The
minor mode can be turned on only if semantic feature is available and
the current buffer was set up for parsing.  Return non-nil if the
minor mode is enabled.

\\{semantic-show-dirty-mode-map}" t nil)

(autoload 'global-semantic-show-unmatched-syntax-mode "semantic-util-modes" "\
Toggle global use of option `semantic-show-unmatched-syntax-mode'.
If ARG is positive, enable, if it is negative, disable.
If ARG is nil, then toggle." t nil)

(defcustom global-semantic-show-unmatched-syntax-mode nil "*If non-nil, enable global use of show-unmatched-syntax mode." :group 'semantic :type 'boolean :require 'semantic-util-modes :initialize 'custom-initialize-default :set (lambda (sym val) (global-semantic-show-unmatched-syntax-mode (if val 1 -1))))

(autoload 'semantic-show-unmatched-syntax-mode "semantic-util-modes" "\
Minor mode to highlight unmatched syntax tokens.
With prefix argument ARG, turn on if positive, otherwise off.  The
minor mode can be turned on only if semantic feature is available and
the current buffer was set up for parsing.  Return non-nil if the
minor mode is enabled.

\\{semantic-show-unmatched-syntax-mode-map}" t nil)

(defcustom global-semantic-auto-parse-mode nil "*If non-nil, enable global use of auto-parse mode." :group 'semantic :type 'boolean :require 'semantic-util-modes :initialize 'custom-initialize-default :set (lambda (sym val) (global-semantic-auto-parse-mode (if val 1 -1))))

(autoload 'global-semantic-auto-parse-mode "semantic-util-modes" "\
Toggle global use of option `semantic-auto-parse-mode'.
If ARG is positive, enable, if it is negative, disable.
If ARG is nil, then toggle." t nil)

(autoload 'semantic-auto-parse-mode "semantic-util-modes" "\
Minor mode to auto parse buffer following a change.
With prefix argument ARG, turn on if positive, otherwise off.  The
minor mode can be turned on only if semantic feature is available and
the current buffer was set up for parsing.  Return non-nil if the
minor mode is enabled." t nil)

(autoload 'global-semantic-summary-mode "semantic-util-modes" "\
Toggle global use of option `semantic-summary-mode'.
If ARG is positive, enable, if it is negative, disable.
If ARG is nil, then toggle." t nil)

(defcustom global-semantic-summary-mode nil "*If non-nil, enable global use of summary mode." :group 'semantic :type 'boolean :require 'semantic-util-modes :initialize 'custom-initialize-default :set (lambda (sym val) (global-semantic-summary-mode (if val 1 -1))))

(autoload 'semantic-summary-mode "semantic-util-modes" "\
Minor mode to show useful things about tokens in echo area.
Enables/disables option `eldoc-mode' which supplies the support functions for
this minor mode.
With prefix argument ARG, turn on if positive, otherwise off.  The
minor mode can be turned on only if semantic feature is available and
the current buffer was set up for parsing.  Return non-nil if the
minor mode is enabled.

\\{semantic-summary-mode-map}" t nil)

;;;***

;;;### (autoloads (semantic-bovinate-toplevel) "semantic" "semantic/semantic.el")

(autoload 'semantic-bovinate-toplevel "semantic" "\
Bovinate the entire current buffer.
If the optional argument CHECKCACHE is non-nil, then make sure the
cached token list is up to date.  If a partial reparse is possible, do
that, otherwise, do a full reparse." nil nil)

;;;***

;;;### (autoloads (global-semanticdb-minor-mode) "semanticdb" "semantic/semanticdb.el")

(defcustom semanticdb-global-mode nil "*If non-nil enable the use of `semanticdb-minor-mode'." :group 'semantic :type 'boolean :require 'semanticdb :initialize 'custom-initialize-default :set (lambda (sym val) (global-semanticdb-minor-mode (if val 1 -1)) (custom-set-default sym val)))

(autoload 'global-semanticdb-minor-mode "semanticdb" "\
Toggle the use of `semanticdb-minor-mode'.
If ARG is positive, enable, if it is negative, disable.
If ARG is nil, then toggle." t nil)

;;;***

;;;### (autoloads (senator-try-expand-semantic global-senator-minor-mode senator-minor-mode senator-word-search-backward senator-re-search-backward senator-search-backward senator-word-search-forward senator-re-search-forward senator-search-forward senator-completion-menu-popup senator-complete-symbol senator-jump-regexp senator-jump senator-previous-token senator-next-token) "senator" "semantic/senator.el")

(defcustom global-senator-minor-mode nil "*If non-nil enable global use of senator minor mode." :group 'senator :type 'boolean :require 'senator :initialize 'custom-initialize-default :set (lambda (sym val) (global-senator-minor-mode (if val 1 -1))))

(autoload 'senator-next-token "senator" "\
Navigate to the next Semantic token.
Return the token or nil if at end of buffer." t nil)

(autoload 'senator-previous-token "senator" "\
Navigate to the previous Semantic token.
Return the token or nil if at beginning of buffer." t nil)

(autoload 'senator-jump "senator" "\
Jump to the semantic symbol SYM.

If optional IN-CONTEXT is non-nil jump in the local type's context
\(see function `senator-current-type-context').  If optional
NO-DEFAULT is non-nil do not provide a default value.

When called interactively you can combine the IN-CONTEXT and
NO-DEFAULT switches like this:

- - \\[universal-argument]       IN-CONTEXT.
- - \\[universal-argument] -     NO-DEFAULT.
- - \\[universal-argument] \\[universal-argument]   IN-CONTEXT + NO-DEFAULT." t nil)

(autoload 'senator-jump-regexp "senator" "\
Jump to the semantic symbol SYMREGEX.
SYMREGEX is treated as a regular expression.

If optional IN-CONTEXT is non-nil jump in the local type's context
\(see function `senator-current-type-context').  If optional
NO-DEFAULT is non-nil do not provide a default value and move to the
next match of SYMREGEX.  NOTE: Doesn't actually work yet.

When called interactively you can combine the IN-CONTEXT and
NO-DEFAULT switches like this:

- - \\[universal-argument]       IN-CONTEXT.
- - \\[universal-argument] -     NO-DEFAULT.
- - \\[universal-argument] \\[universal-argument]   IN-CONTEXT + NO-DEFAULT." t nil)

(autoload 'senator-complete-symbol "senator" "\
Complete the current symbol under point.
If optional argument CYCLE-ONCE is non-nil, only cycle through the list
of completions once, doing nothing where there are no more matches." t nil)

(autoload 'senator-completion-menu-popup "senator" "\
Popup a completion menu for the symbol at point.
The popup menu displays all of the possible completions for the symbol
it was invoked on.  To automatically split large menus this function
use `imenu--mouse-menu' to handle the popup menu." t nil)

(autoload 'senator-search-forward "senator" "\
Search semantic tokens forward from point for string WHAT.
Set point to the end of the occurrence found, and return point.  See
`search-forward' for details and the meaning of BOUND NOERROR and
COUNT.  COUNT is just ignored in the current implementation." t nil)

(autoload 'senator-re-search-forward "senator" "\
Search semantic tokens forward from point for regexp WHAT.
Set point to the end of the occurrence found, and return point.  See
`re-search-forward' for details and the meaning of BOUND NOERROR and
COUNT.  COUNT is just ignored in the current implementation." t nil)

(autoload 'senator-word-search-forward "senator" "\
Search semantic tokens forward from point for word WHAT.
Set point to the end of the occurrence found, and return point.  See
`word-search-forward' for details and the meaning of BOUND NOERROR and
COUNT.  COUNT is just ignored in the current implementation." t nil)

(autoload 'senator-search-backward "senator" "\
Search semantic tokens backward from point for string WHAT.
Set point to the beginning of the occurrence found, and return point.
See `search-backward' for details and the meaning of BOUND NOERROR and
COUNT.  COUNT is just ignored in the current implementation." t nil)

(autoload 'senator-re-search-backward "senator" "\
Search semantic tokens backward from point for regexp WHAT.
Set point to the beginning of the occurrence found, and return point.
See `re-search-backward' for details and the meaning of BOUND NOERROR
and COUNT.  COUNT is just ignored in the current implementation." t nil)

(autoload 'senator-word-search-backward "senator" "\
Search semantic tokens backward from point for word WHAT.
Set point to the beginning of the occurrence found, and return point.
See `word-search-backward' for details and the meaning of BOUND
NOERROR and COUNT.  COUNT is just ignored in the current
implementation." t nil)

(autoload 'senator-minor-mode "senator" "\
Toggle senator minor mode.
With prefix argument ARG, turn on if positive, otherwise off.  The
minor mode can be turned on only if semantic feature is available and
the current buffer was set up for parsing.  Return non-nil if the
minor mode is enabled.

\\{senator-mode-map}" t nil)

(autoload 'global-senator-minor-mode "senator" "\
Toggle global use of senator minor mode.
If ARG is positive, enable, if it is negative, disable.
If ARG is nil, then toggle." t nil)

(autoload 'senator-try-expand-semantic "senator" "\
Attempt inline completion at the cursor.
Use Semantic, or the semantic database to look up possible
completions.  The argument OLD has to be nil the first call of this
function.  It returns t if a unique, possibly partial, completion is
found, nil otherwise." nil nil)

;;;***

(provide 'semantic-autoloads)
